#!@PERL@

# --8<--8<--8<--8<--
#
# Copyright (C) 2010 Smithsonian Astrophysical Observatory
#
# This file is part of make-world
#
# make-world is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# make-world is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#       Free Software Foundation, Inc.
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--


use strict;
use warnings;

package MakeWorld::Log;

use DateTime::Tiny;

use MakeWorld::IORedir;

use Data::Dump qw( dump );
use Scope::Guard 'guard';
use Safe::Isa;

use base qw( Exporter );

use constant Fatal  => -1;
use constant Error  => 0;
use constant Warn   => 1;
use constant Notice => 2;
use constant Info   => 3;
use constant Debug  => 4;

our @EXPORT = qw(
  log_error
  log_fatal
  log_warn
  log_notice
  log_info
  log_debug
  log_pfx
  log_lvl
  log_timedate
  log_level
  log_at_level
);

my $log_level = Warn;

my @prefix;
my $prefix = '';

my $prefix_timedate = 0;

sub new {

    my $class = shift;
    push_pfx(@prefix, shift);

    return bless {}, $class;
}

sub DESTROY {
    &pop_pfx;
}

sub mkpfx
{
    $prefix = join( '', map { "$_: " } @prefix );
}

sub push_pfx
{
    push @prefix, shift;
    mkpfx();
}

sub pop_pfx
{
    pop @prefix;
    mkpfx();
}

sub log_level
{
    my $old = $log_level;
    $log_level = $_[0] if @_;
    return $old;
}

sub log_timedate {
    my $old = $prefix_timedate;
    $prefix_timedate = $_[0] if @_;
    return $old;
}

sub log_error {

    unshift @_, Error;
    &log_at_level;
}

sub log_fatal {

    unshift @_, Fatal;
    die &log_at_level;
}

sub log_warn {

    unshift @_, Warn;
    &log_at_level;
}

sub log_notice {

    unshift @_, Notice;
    &log_at_level;
}

sub log_info {

    unshift @_, Info;
    &log_at_level;
}

sub log_debug {

    unshift @_, Debug;
    &log_at_level;
}

sub log_at_level {
    my $msg_level = shift;

    my $guard = shift if $_[0]->$_isa( 'Scope::Guard' );

    if ( $msg_level <= $log_level ) {
        my $rd = MakeWorld::IORedir->new( reset => 1 );

        my @lines;

        for my $msg ( @_ ) {
            if ( !ref $msg ) {
                push @lines, split /\n/, $msg;
            }
            elsif ( 'ARRAY' eq ref $msg ) {
                push @lines, map { "[$_]" } @$msg;
            }
            elsif ( 'HASH' eq ref $msg ) {
                push @lines, split /\n/, dump( $msg );
            }

        }

        chomp( @lines );
        my $lprefix = join( '',
            (
                $prefix_timedate ? ( DateTime::Tiny->now->as_string, ': ' ) : ()
            ),
            $prefix );

        my $msg =
            $lprefix .
            join( "\n$lprefix", @lines ) .
            "\n";

        print $msg;

        return $prefix . join( "\n$prefix", @lines ) . "\n"
          if $msg_level == Fatal;
    }
}


sub log_pfx {

    my ( $prefix ) = @_;

    &push_pfx( $prefix );

    return guard { &pop_pfx };
}

sub log_lvl {

    my ( $level ) = @_;

    my $old_level = log_level( $level );

    return guard { $log_level = $old_level  };
}


1;
