package MakeWorld::Options;

use strict;
use warnings;

use Moo;

use Getopt::Long qw[ GetOptionsFromArray ];
use Getopt::ArgvFile qw[argvFile];

use File::Path 'make_path';

use MakeWorld::Log;

use File::Spec::Functions qw/ catfile catdir rel2abs /;

my %Option;

sub opt(@) {

    my ( $option, %attr ) = @_;

    my $spec = delete $attr{spec};
    my $alt  = delete $attr{alt};

    $Option{$option}{spec} = join( '',
        $option,
        ( defined $alt ? ( '|', $alt ) : () ),
        ( defined $spec ? $spec : () ) );

    has $option => %attr;
}


opt builddir => (
    is      => 'ro',
    lazy    => 1,
    default => '.',
    coerce  => sub { rel2abs( $_[0] ) },
    spec    => '=s',
);

opt cache => (
    is        => 'ro',
    predicate => 1,
    spec      => '=s',
);

opt clean => (
    is      => 'ro',
    default => 1,
    spec    => '!'
);

opt create_tld => (
    is      => 'ro',
    default => 0,
);

opt dryrun => (
    is      => 'ro',
    default => 0,
    alt     => 'n',
    spec    => '+',
);

opt define => (
    is      => 'ro',
    default => sub { {} },
    spec    => '=s%',
    alt     => 'D',
);

opt odefine => (
    is      => 'ro',
    default => sub { {} },
    spec    => '=s%',
);

opt exclude => (
    is      => 'ro',
    lazy    => 1,
    default => sub { {} },

    # change excluded packages list to a hash to make
    # life easier split on commas while we're at it
    coerce => sub {
        return $_[0] if 'HASH' eq ref $_[0];
        return {
            map { $_ => 1 }
            map { split( /[,]/, $_ ) } @{ $_[0] } };
    },
    spec => '=s@',
);

opt exec_prefix => (
    is      => 'ro',
    lazy    => 1,
    default => sub { $_[0]->prefix },
    coerce  => sub { rel2abs( $_[0] ) },
    spec    => '=s',
);

opt force => (
    is      => 'ro',
    default => 0,
    spec    => '!',
);

opt help => (
    is      => 'ro',
    default => 0,
);

opt logfile => (
    is        => 'ro',
    predicate => 1,
    clearer => 1,
    spec      => ':s',
);

opt logredir => (
    is      => 'ro',
    default => 0,
    spec    => '!',
);

opt make_cmd => (
    is      => 'ro',
    default => 'make',
    spec    => '=s',
);

opt nopkgs => (
    is      => 'ro',
    default => 0,
);

opt only => (
    is      => 'ro',
    default => 0,
);

has packages => (
    is      => 'ro',
    default => sub { [] } );

opt patch_cmd => (
    is      => 'ro',
    default => 'patch',
    spec    => '=s',
);

opt pkgs => (
    is      => 'ro',
    lazy    => 1,
    default => sub { [] },
    spec    => '=s@',
);

opt prefix => (
    is        => 'ro',
    lazy      => 1,
    coerce    => sub { rel2abs( $_[0] ) },
    predicate => 1,
    spec      => '=s',
);

opt quiet => (
    is      => 'ro',
    default => 0
);

opt srcdir => (
    is      => 'ro',
    lazy    => 1,
    default => sub { [] },
    coerce  => sub {
        [ map { rel2abs( $_ ) } @{ $_[0] } ];
    },
    spec => '=s@',
);

opt stages => (
    is      => 'ro',
    default => sub { [] },
    lazy    => 1,
    coerce  => sub {
        return $_[0] if 'ARRAY' eq ref $_[0];
        return [ split( ',', $_[0] ) ];
    },
    spec => '=s',
);

opt statusfile => (
    is      => 'ro',
    lazy    => 1,
    default => sub {
        catfile( $_[0]->builddir, 'make-world.status' );
    },
    coerce => sub { rel2abs( $_[0] ) },
    spec   => '=s',
);

opt manual => (
    is      => 'ro',
    default => 0,
);

opt verbose => (
    is      => 'rwp',
    default => 2,
    coerce  => sub { $_[0] + 1 },
    spec    => ':+',
);

opt version => (
    is      => 'ro',
    default => 0,
);

opt defoption => (
    is      => 'ro',
    default => sub { {} },
    spec    => '=s%',
);

opt option => (
    is      => 'ro',
    default => sub { {} },
    spec    => '=s%',
);

has PERL5LIB => (
    is      => 'ro',
    lazy    => 1,
    default => sub { [ catdir( $_[0]->prefix, qw{ lib perl5 } ) ] },
);

has progress_bar => (
    is        => 'rwp',
    init_args => undef,
    default   => 0,
);

has argv => (
             is => 'ro',
            );

sub BUILDARGS {

    my $self = shift;

    my $argv = @_ ? [@_] : \@ARGV;

    argvFile( resolveEnvVars => 1 );

    my %opt;

    $opt{argv} = [ @$argv ];

    my @spec = ( map $Option{$_}{spec}, keys %Option );

    eval {
        local $SIG{__WARN__} = sub { die $_[0] };

        GetOptionsFromArray( $argv, \%opt, @spec );
    };

    die $@ if $@;

    $opt{packages} = [@ARGV];

    return \%opt;
}

sub BUILD {

    my $self = shift;

    return if $self->version || $self->help || $self->manual;

    $self->_set_progress_bar( 1 ) if $self->has_logfile && !$self->quiet;

    $self->_set_verbose( 0 ) if $self->quiet && !$self->has_logfile;

    $self->clear_logfile if $self->has_logfile && ! length( $self->logfile );

    log_level( $self->verbose );

    die( "--srcdir not specified\n" )
      unless @{ $self->srcdir };

    die( "--prefix not specified\n" )
      unless $self->has_prefix;

}

1;
