# --8<--8<--8<--8<--
#
# Copyright (C) 2010 Smithsonian Astrophysical Observatory
#
# This file is part of make-world
#
# make-world is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# make-world is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#       Free Software Foundation, Inc.
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--


package MakeWorld::PackageSpec;

use strict;
use warnings;

use base qw[ Exporter ];

our @EXPORT_OK = qw[ load_pkgfile ];

use Module::Load qw[ load ];

use XML::Tiny qw[ parsefile ];

use MakeWorld::Log;
use MakeWorld::Global;
use MakeWorld::Package;
use MakeWorld::Interpolate;
use MakeWorld::Run;
use MakeWorld::VarStack;
use MakeWorld::Check;
use MakeWorld::Mogrify;
use MakeWorld::Utils qw[ find_exe ];

my %ConditionalVar;

use constant {
    T_COND   => 0,
    T_DELETE => 1,
    T_RETURN => 2
};

# load package descriptions from package files
sub load_pkgfile
{
    my ( $pkgfiles ) = @_;

    my @pkgs;

    for my $pkgfile ( @$pkgfiles )
    {
	log_info( "Loading package file $pkgfile" );

	my $specs = parsefile( $pkgfile );

	my $element = $specs->[0]{name};
	die( "expected top level <packages> element; got $element\n" )
	  unless $element eq 'packages';

	die( "illegal element type: $element\n" )
	  unless $specs->[0]{type} eq 'e';

	for my $pkgspec ( map { MakeWorld::PackageSpec->new( $_ ) }
			           @{ $specs->[0]{content} } )
	{
	    die( "expected <package> element, got <", $pkgspec->{name}, ">\n" )
	      unless $pkgspec->{name} eq 'package';


	    die( "unexpected content in <package> element: $pkgspec->{content}\n" )
	      unless $pkgspec->{type} eq 'e';

	    # need the top level package attributes (such as matchre, versionre)
	    # before the actual parsing gets done.
	    $pkgspec->fix_attr;

	    my $pkg_name = $pkgspec->attr->{name}
	      or die( qq{<package> must have a "name" attribute\n} );

	    push @pkgs, MakeWorld::Package->new( $pkg_name,
						 pkgspec => $pkgspec,
						 option => $pkgspec->attr );
	}
    }

    return @pkgs;
}

#------------------------------------------------------------------------


sub new {

    my ( $class, $node ) = @_;

    return bless $node, $class;
}

#------------------------------------------------------------------------

sub attr {  $_[0]->{attrib} }

sub type { $_[0]->{type} }

sub text { $_[0]->attr->{_text} }

sub has_text { exists $_[0]->attr->{_text} && length( $_[0]->attr->{_text} ) }

sub content { $_[0]->{content} }

sub has_content { @{$_[0]->{content}} > 0 }

sub is_conditional { exists $_[0]->{name} && $_[0]->{name} =~ /^if/ }

# does the node of any attributes, exluding _text
sub has_attr { keys(%{$_[0]->attr}) - ( exists($_[0]->attr->{_text}) ? 1: 0 ) }

sub attr_from_text {

    my ( $node, $var ) = @_;

    if ( $node->has_text )
    {
	die( "<$node->{name}>: specify $var either as an attribute or as text, but not both\n" )
	  if exists $node->attr->{$var};

	$node->attr->{$var} = delete $node->attr->{_text};
    }

    return;
}

sub fix_attr {

    my ( $self ) = @_;

    my $attr = $self->attr;

    for my $key ( keys %{ $attr } )
    {
	my $value = $attr->{$key};
	$value = 1 if lc($value) =~ /\A(?:1|true|on|yes)\z/;
	$value = 0 if lc($value) =~ /\A(?:0|false|off|no)\z/;
	$attr->{$key} = $key eq 'arg' ? [ $value ] : $value;
    }

    return $self->attr;
}

sub merge_attr {

    my ( $self, $import ) = @_;

    my $attr = $self->attr;

    log_debug( log_pfx( 'attr'), $attr );
    log_debug( log_pfx( 'import'), $import );

    while ( my ( $key, $val ) = each %$import )
    {
	# if $val is a hash and its only key is _text, it's the result of
	#  <element>text</element>
	# turn that into an attribute element="text"

	$val = $val->{_text}
	  if 'HASH' eq ref $val && keys(%$val) == 1 && exists $val->{_text};

	# if the key exists and it's an array, push the value onto it
	if ( exists $attr->{$key} && 'ARRAY' eq ref $attr->{$key} )
	{
	    push @{$attr->{$key}}, 'ARRAY' eq ref $val ? @$val : $val;
	}

	# _text values are concatenated
	elsif ( $key eq '_text' )
	{
	    if(  exists ( $attr->{_text} ) )
	    {
		$attr->{_text} .= $val;
	    }
	    else
	    {
		$attr->{_text} = $val;
	    }
	}

	# if it's an <arg> value, make sure it's an array,
	elsif ( $key eq 'arg' )
	{
	    $attr->{$key} = 'ARRAY' eq ref $val ? $val : [ $val ];
	}

	# everything else just overwrites the attribute
	else
	{
	    $attr->{$key} = $val;
	}
    }

    log_debug( log_pfx( 'attr'), $attr );

    return;
}

sub check {

    my ( $self, %tpl ) = @_;

    # _text is handled differently, as the user doesn't specify it as
    # an attribute
    my $has_text = $self->has_text;
    my $wants_text = exists $tpl{_text};
    my $req_text = $wants_text && exists $tpl{_text}{required} && $tpl{_text}{required};

    die( "<$self->{name}> requires text\n" )
      if ( $req_text && ! $has_text );

    die( "<$self->{name}> must not have text\n" )
      if ( $has_text && ! $wants_text );

    # everything else is as usual; for some reason can't get lvalue to work here
    $self->{attrib} = eval{ MakeWorld::Check::check( \%tpl, $self->attr ) }
      or die( "<$self->{name}>: $@\n" );

    return;
}

#------------------------------------------------------------------------

sub parse {

    my ( $self, $pkg ) = @_;

    my $stack = MakeWorld::VarStack->new;

    return $self->parse_node( $stack, $pkg );
}


sub parse_node {

    my ( $node, $stack, $pkg ) = @_;

    $node->fix_attr;

    $stack = $stack->new;

    my $name = $node->{name};

    # convert text to an attribute and return it; nothing else to do
    if ( $node->type eq 't' )
    {
	return ( _text => $node->content );
    }

    # conditionals are special; don't descend into children nodes
    if ( $node->is_conditional )
    {
	die( "<$name>: unknown conditional\n" )
	  unless my $sub = $node->can( "_$name" );

	my ( $type, @ret ) = $sub->( $node, $stack, $pkg );

	die( "<$name>: internal error: expected conditional\n" )
	  unless $type eq T_COND;

	return unless $ret[0];

    }

    # now, process any sub elements, as they may modify this one.
    for my $sn ( @{ $node->content } )
    {
	my ($name, $res )
	  = MakeWorld::PackageSpec->new($sn)->parse_node( $stack, $pkg );

	# nodes may be removed from the tree by parse_node; it returns
	# $res = undef in that instance.
	next unless defined $res;

	# node contents may be promoted up a level (e.g. a conditional
	# node's contents); parse_node returns $name = undef in this
	# case.
	$node->merge_attr( defined $name ? { $name => $res } : $res );
    }

    # now process the possibly modified node

    # a condition succeeded, promote the children
    if ( $node->is_conditional )
    {
	# remove attributes that shouldn't get passed up
	delete @{$node->attr}{qw[ expr var ]};
	return( undef, $node->attr );
    }

    # is this an immediate directive?
    elsif ( my $sub = $node->can( "_$name" ) )
    {
	my ( $type, @ret ) = $sub->( $node, $stack, $pkg );

	if ( $type eq T_COND )
	{
	    die( "internal error: unexpectedly got T_COND\n" );
	}

	elsif ( $type eq T_DELETE )
	{
	    return;
	}

	elsif ( $type eq T_RETURN )
	{
	    return @ret;
	}

	else
	{
	    die( "internal error; unknown returned type for $name\n" );
	}

    }

    # is this a package mogrification command? inserted into the package hooks list,
    # and remove from the node tree
    elsif ( my $class = MakeWorld::Mogrify::is_cmd( $name ) )
    {
	load $class;
	my $attr = eval { $class->verify( $node->attr, 1 ) }
	  or die "<$name>: $@\n";
	$node->{attrib} = $attr;

	$pkg->add_mogrify( { hook => $attr->{hook},
			     name => $name,
			     class => $class,
			     attr => $node->attr },
			 );
	return;
    }

    # what's left?  non-function nodes. promote them as attributes to the parent element
    return ( $name, $node->attr );
}


#------------------------------------------------------------------------

# handle <arg> elements differently.  they can occur more than
# once and thus get bundled into arrays.

sub _arg {

    my ( $node, $stack, $pkg ) = @_;

    die( "<arg> must not have attributes\n" )
      if $node->has_attr;

    die ("<arg> must have text\n" )
      unless $node->has_text;

    my $res = $node->text;
    $res =~ s/^\s+//;
    $res =~ s/\s+$//;
    return ( T_RETURN, arg => $res );
}

sub _env {

    my ( $node, $stack, $pkg ) = @_;

    # attributes may only be strings
    my @not_strings = grep { ref $node->attr->{$_} } keys %{$node->attr};
    die( "<env> may only have string attributes. these aren't: ", 

	 join( ', ', @not_strings), "\n")
      if @not_strings;

    # <env> doesn't do text
    die( "<env> may not have text\n" )
      if $node->has_text;

    my $attr = $node->attr;

    log_debug( log_pfx( '<env>' ), $attr );

    $pkg->set_env( $attr );

    return T_DELETE;
}


# execute a command and stash the results

sub _run {

    my ( $node, $stack, $pkg ) = @_;

    $node->check( exe    => { required => 1 },
		  arg    => { default => [] },
		  exit   => { default => 'exit' },
		  stdout => { default => 'stdout' },
		  stderr => { default => 'stderr' },
		);

    my $attr = $node->attr;
    my @keys = @{$attr}{ qw[ exit stdout stderr ] };

    $node = interpolate( $attr, $pkg->archive );

    my %res;
    my $file = eval { find_exe( $attr->{exe} ) };

    die( "<run>: ", $attr->{exe}, ": non-existent or non-executable file\n" )
	unless defined $file;

    @res{@keys} = capture( $file, $attr->{arg} );

    $stack->parent->set( %res );
    $stack->parent->dump( @keys );

    return T_DELETE;
}

sub __safe_eval {

    my ( $node, $stack, $expr ) = @_;

    require Safe;
    my $cpt = Safe->new;
    $cpt->permit( ':browse' );

    for my $vars ( $stack->vars, \%MakeWorld::Interpolate::InterpolateVars )
    {
	${$cpt->varglob($_)} = $vars->{$_}
	  foreach keys %$vars;
    }

    %{$cpt->varglob('ENV')} = %ENV;

    my $res = $cpt->reval( $expr ) || 0;
    die( "<eval>: error in expression: $expr\n$@\n" )
      if $@;

    log_debug( "$node->{name}: ($expr) = $res\n" );

    return $res;
}

sub _if {

    my ( $node, $stack, $pkg ) = @_;

    $node->check( expr => { required => 1 } );

    return T_COND, $node->__safe_eval( $stack, $node->attr->{expr} );
}

sub _iftrue {

    my ( $node, $stack, $pkg ) = @_;

    $node->check( var => { default => '%eval' } );

    my $var = $node->attr->{var};

    die( "<iftrue>: no variable with the name $var\n" )
      unless $stack->exists($var);

    return T_COND, $stack->value($var);
}

sub _iffalse {

    my ( $node, $stack, $pkg ) = @_;

    $node->check( var => { default => '%eval' } );

    my $var = $node->attr->{var};

    die( "<iffalse>: no variable with the name $var\n" )
      unless $stack->exists($var);

    return T_COND, ! $stack->value($var);
}

sub _eval {

    my ( $node, $stack, $pkg ) = @_;

    $node->attr_from_text( 'expr' );

    $node->check( expr => { required => 1 },
		  var  => { default => '%eval' },
		);

    my $attr = $node->attr;

    my $res = $node->__safe_eval( $stack, $attr->{expr} );

    $stack->parent->set( $attr->{var} => $res );
    $stack->parent->dump( $attr->{var} );

    return T_DELETE;
}

# read a file into a variable

sub _read {

    my ( $node, $stack, $pkg ) = @_;

    $node->check( file => { required => 1 },
		  var  => { default => '_buffer' },
		);

    my $attr = $node->attr;
    log_debug( log_pfx( '<read>' ), $attr );

    $attr = interpolate( $attr, $pkg->archive );

    require IO::File;

    my $fh = IO::File->new( $attr->{file}, 'r' )
      or die( "<read>: error opening $attr->{file}\n" );

    my $buffer = eval { local $/ = undef; $fh->getline };

    $stack->parent->set( $attr->{var} => $buffer );
    $stack->parent->dump( $attr->{var} );


    return T_DELETE;
}

sub _debug {

    my ( $node, $stack, $pkg ) = @_;

    $node->attr_from_text( 'str' );

    $node->check( str => { required => 1 } );

    log_debug( $node->attr->{str} );

    return T_DELETE;
}

1;
