package MakeWorld::Stages;

use strict;
use warnings;

use MakeWorld::Global;
use MakeWorld::Check qw[ check ];

# Stage phases
my @Phases = qw[ pre main post ];

my @LegalStages = qw[ configure build test install ] ;
my %LegalStages = map { $_ => 1 } @LegalStages;


sub _split_stage { split( /[,\s]/, $_[0] ) }

sub legal_stage{

    return 0 == grep { ! defined $LegalStages{$_} } _split_stage( $_[0] );

}
sub stagedef{ defined $_[0] ? [ _split_stage( $_[0] ) ] : \@LegalStages }

#----------------------------------------------------------------#

sub new {

    my ( $class, %opts ) = @_;

    my $self = check( { stages => { strict_type => 1,
				    default => \@LegalStages },
			filter => { required => 0 },
		      },
		      \%opts
		    );

    $self->{stage} = { map { $_ => 1 } 'extract', @{$self->{stages}} };

    $self->{components} = [
	'extract', 'extract:post',
	map { my $stage = $_;
	      $stage, map { "$stage:$_" } @Phases
	} @{$self->{stages}}
	];

    $self->{component} = { map { $_ => 1 } @{$self->{components}} };
    $self->{allowed_stage} = { %{$self->{component}} };

    $self->{hook} = { map { ( "$_:pre" => 1, "$_:post" => 1 ) }
		      @{ $self->{stages} } };

    $self->{hook}{'extract:post'} = 1;

    bless $self, $class;

    $self->allow_stages( $self->{filter} );

    return $self;
}


################################################################################

sub _is_component { exists $_[0]->{component}{$_[1]} }

sub is_stage { exists $_[0]->{stage}{$_[1]} }

sub last_stage { $_[0]->{stages}[-1] }

sub is_legal_hook {exists $_[0]->{hook}{$_[1]} }

################################################################################

sub allow_stages {

    my ( $self, $filter ) = @_;

    # @filter is a list of included and excluded stages.

    # If the first element is an inclusion, then the initial set of
    # stages is empty.

    # If the first element is a deletion, then the initial set of
    # stages is the possible set of stages.

    my @filter = 'ARRAY' eq ref $filter   ? @$filter
               : defined $filter          ? _split_stage( $filter )
		                          : ();

    my %allowed = @filter == 0 || $filter[0] =~ /^-/
                ? %{ $self->{allowed_stage} }
		: ();

    # if run is defined, only the provided stages are allowed
    for my $filter (@filter) {

	my ( $exclude, $comp ) = $filter =~ /(-)?(.*)/;

        die( "unrecognized stage '$comp'\n" )
          if !  $exclude && ! $self->_is_component( $comp );

	if ( defined $exclude ) {

	    # if it's just the stage name, disallow all of the phases
	    delete @allowed{
		defined $self->is_stage($comp) ? ( $comp, map { "$comp:$_" } @Phases ) : $comp
	    };

	} else {

	    $allowed{$comp} = 1;

	    # if it's just the stage name, allow all of the phases
	    do { $allowed{ "$comp:$_" } = 1 foreach @Phases }
		if $self->is_stage( $comp );

	}

    }

    $self->{allowed_stage} = \%allowed;

    return;
}

#----------------------------------------------------------------#

sub parse_hook { shift; return split(':', $_[0] ) }

sub ok_to_run {
    my $self = shift;
    exists $self->{allowed_stage}{ join(':', @_ ) };
}

1;
